@@ -1,7 +1,8 @@
 import re
 from collections import OrderedDict
+from datetime import datetime
+from typing import Union
 
-from django.conf import settings
 from redis.exceptions import ConnectionError
 
 from ..exceptions import ConnectionInterrupted
@@ -22,10 +23,10 @@ class ShardClient(DefaultClient):
         self._ring = HashRing(self._server)
         self._serverdict = self.connect()
 
-    def get_client(self, write=True):
+    def get_client(self, *args, **kwargs):
         raise NotImplementedError
 
-    def connect(self):
+    def connect(self, index=0):
         connection_dict = {}
         for name in self._server:
             connection_dict[name] = self.connection_factory.connect(name)
@@ -137,6 +138,18 @@ class ShardClient(DefaultClient):
 
         return super().ttl(key=key, version=version, client=client)
 
+    def pttl(self, key, version=None, client=None):
+        """
+        Executes PTTL redis command and return the "time-to-live" of specified key
+        in milliseconds. If key is a non volatile key, it returns None.
+        """
+
+        if client is None:
+            key = self.make_key(key, version=version)
+            client = self.get_server(key)
+
+        return super().pttl(key=key, version=version, client=client)
+
     def persist(self, key, version=None, client=None):
         if client is None:
             key = self.make_key(key, version=version)
@@ -151,6 +164,25 @@ class ShardClient(DefaultClient):
 
         return super().expire(key=key, timeout=timeout, version=version, client=client)
 
+    def pexpire(self, key, timeout, version=None, client=None):
+        if client is None:
+            key = self.make_key(key, version=version)
+            client = self.get_server(key)
+
+        return super().pexpire(key=key, timeout=timeout, version=version, client=client)
+
+    def expire_at(self, key, when: Union[datetime, int], version=None, client=None):
+        """
+        Set an expire flag on a ``key`` to ``when`` on a shard client.
+        ``when`` which can be represented as an integer indicating unix
+        time or a Python datetime object.
+        """
+        if client is None:
+            key = self.make_key(key, version=version)
+            client = self.get_server(key)
+
+        return super().expire_at(key=key, when=when, version=version, client=client)
+
     def lock(
         self,
         key,
@@ -198,7 +230,7 @@ class ShardClient(DefaultClient):
         value = self.get(old_key, version=version, client=client)
 
         try:
-            ttl = client.ttl(old_key)
+            ttl = self.ttl(old_key, version=version, client=client)
         except ConnectionError as e:
             raise ConnectionInterrupted(connection=client) from e
 
@@ -265,11 +297,9 @@ class ShardClient(DefaultClient):
                 res += connection.delete(*keys)
         return res
 
-    def close(self, **kwargs):
-        if getattr(settings, "DJANGO_REDIS_CLOSE_CONNECTION", False):
-            for client in self._serverdict.values():
-                for c in client.connection_pool._available_connections:
-                    c.disconnect()
+    def do_close_clients(self):
+        for client in self._serverdict.values():
+            self.disconnect(client=client)
 
     def touch(self, key, timeout=DEFAULT_TIMEOUT, version=None, client=None):
         if client is None:
@@ -277,3 +307,7 @@ class ShardClient(DefaultClient):
             client = self.get_server(key)
 
         return super().touch(key=key, timeout=timeout, version=version, client=client)
+
+    def clear(self, client=None):
+        for connection in self._serverdict.values():
+            connection.flushdb()
