@@ -24,7 +24,7 @@ herd.CACHE_HERD_TIMEOUT = 2
 
 
 def make_key(key, prefix, version):
-    return "{}#{}#{}".format(prefix, version, key)
+    return f"{prefix}#{version}#{key}"
 
 
 def reverse_key(key):
@@ -38,7 +38,9 @@ class DjangoRedisConnectionStrings(unittest.TestCase):
             "redis://localhost/2",
             "rediss://localhost:3333?db=2",
         ]
-        cf = pool.get_connection_factory(options={})
+        cf = pool.get_connection_factory(
+            path="django_redis.pool.ConnectionFactory", options={}
+        )
         for connection_string in connection_strings:
             with self.subTest(connection_string):
                 res = cf.make_connection_params(connection_string)
@@ -55,15 +57,11 @@ class DjangoRedisCacheTestEscapePrefix(unittest.TestCase):
         self.addCleanup(cm.disable)
 
         self.cache = caches["default"]
-        try:
-            self.cache.clear()
-        except Exception:
-            pass
         self.other = caches["with_prefix"]
-        try:
-            self.other.clear()
-        except Exception:
-            pass
+
+    def tearDown(self):
+        self.cache.clear()
+        self.other.clear()
 
     def test_delete_pattern(self):
         self.cache.set("a", "1")
@@ -98,10 +96,9 @@ class DjangoRedisCacheTestCustomKeyFunction(unittest.TestCase):
         self.addCleanup(cm.disable)
 
         self.cache = caches["default"]
-        try:
-            self.cache.clear()
-        except Exception:
-            pass
+
+    def tearDown(self):
+        self.cache.clear()
 
     def test_custom_key_function(self):
         if isinstance(self.cache.client, ShardClient):
@@ -126,10 +123,8 @@ class DjangoRedisCacheTests(unittest.TestCase):
     def setUp(self):
         self.cache = cache
 
-        try:
-            self.cache.clear()
-        except Exception:
-            pass
+    def tearDown(self):
+        self.cache.clear()
 
     def test_setnx(self):
         # we should ensure there is no test_key_nx in redis
@@ -288,7 +283,6 @@ class DjangoRedisCacheTests(unittest.TestCase):
 
         res = cache.get("add_key")
         self.assertEqual(res, "Initial value")
-
         res = self.cache.add("other_key", "New value")
         self.assertIs(res, True)
 
@@ -322,14 +316,12 @@ class DjangoRedisCacheTests(unittest.TestCase):
         value = "value"
 
         with patch.object(pipeline, "set") as mocked_set:
-            self.cache.set(
-                key, value, client=pipeline,
-            )
+            self.cache.set(key, value, client=pipeline)
 
         if isinstance(self.cache.client, herd.HerdClient):
             default_timeout = self.cache.client._backend.default_timeout
             herd_timeout = (default_timeout + herd.CACHE_HERD_TIMEOUT) * 1000
-            herd_pack_value = self.cache.client._pack(value, default_timeout,)
+            herd_pack_value = self.cache.client._pack(value, default_timeout)
             mocked_set.assert_called_once_with(
                 self.cache.client.make_key(key, version=None),
                 self.cache.client.encode(herd_pack_value),
@@ -357,6 +349,28 @@ class DjangoRedisCacheTests(unittest.TestCase):
         res = self.cache.delete("a")
         self.assertFalse(bool(res))
 
+    @patch("django_redis.cache.DJANGO_VERSION", (3, 1, 0, "final", 0))
+    def test_delete_return_value_type_new31(self):
+        """delete() returns a boolean instead of int since django version 3.1"""
+        self.cache.set("a", 1)
+        res = self.cache.delete("a")
+        self.assertEqual(type(res), bool)
+        self.assertTrue(res)
+        res = self.cache.delete("b")
+        self.assertEqual(type(res), bool)
+        self.assertFalse(res)
+
+    @patch("django_redis.cache.DJANGO_VERSION", (3, 0, 1, "final", 0))
+    def test_delete_return_value_type_before31(self):
+        """delete() returns a int before django version 3.1"""
+        self.cache.set("a", 1)
+        res = self.cache.delete("a")
+        self.assertEqual(type(res), int)
+        self.assertEqual(res, 1)
+        res = self.cache.delete("b")
+        self.assertEqual(type(res), int)
+        self.assertEqual(res, 0)
+
     def test_delete_many(self):
         self.cache.set_many({"a": 1, "b": 2, "c": 3})
         res = self.cache.delete_many(["a", "b"])
@@ -414,6 +428,37 @@ class DjangoRedisCacheTests(unittest.TestCase):
         res = self.cache.get("num")
         self.assertEqual(res, 5)
 
+    def test_incr_no_timeout(self):
+        if isinstance(self.cache.client, herd.HerdClient):
+            self.skipTest("HerdClient doesn't support incr")
+
+        self.cache.set("num", 1, timeout=None)
+
+        self.cache.incr("num")
+        res = self.cache.get("num")
+        self.assertEqual(res, 2)
+
+        self.cache.incr("num", 10)
+        res = self.cache.get("num")
+        self.assertEqual(res, 12)
+
+        # max 64 bit signed int
+        self.cache.set("num", 9223372036854775807, timeout=None)
+
+        self.cache.incr("num")
+        res = self.cache.get("num")
+        self.assertEqual(res, 9223372036854775808)
+
+        self.cache.incr("num", 2)
+        res = self.cache.get("num")
+        self.assertEqual(res, 9223372036854775810)
+
+        self.cache.set("num", 3, timeout=None)
+
+        self.cache.incr("num", 2)
+        res = self.cache.get("num")
+        self.assertEqual(res, 5)
+
     def test_incr_error(self):
         if isinstance(self.cache.client, herd.HerdClient):
             self.skipTest("HerdClient doesn't support incr")
@@ -527,6 +572,15 @@ class DjangoRedisCacheTests(unittest.TestCase):
         res = self.cache.get("keytest", version=2)
         self.assertEqual(res, 2)
 
+    def test_ttl_incr_version_no_timeout(self):
+        self.cache.set("my_key", "hello world!", timeout=None)
+
+        self.cache.incr_version("my_key")
+
+        my_value = self.cache.get("my_key", version=2)
+
+        self.assertEqual(my_value, "hello world!")
+
     def test_delete_pattern(self):
         for key in ["foo-aa", "foo-ab", "foo-bb", "foo-bc"]:
             self.cache.set(key, "foo")
@@ -561,35 +615,72 @@ class DjangoRedisCacheTests(unittest.TestCase):
             "*foo-a*", itersize=expected_count
         )
 
+    @override_settings(DJANGO_REDIS_CLOSE_CONNECTION=True)
     def test_close(self):
-        cache = caches["default"]
-        cache.set("f", "1")
-        cache.close()
+        self.cache.set("f", "1")
+        self.cache.close()
+
+    def test_close_client(self):
+        with patch.object(self.cache.client, "close") as mock:
+            self.cache.close()
+            assert mock.called
 
     def test_ttl(self):
-        cache = caches["default"]
+        self.cache.set("foo", "bar", 10)
+        ttl = self.cache.ttl("foo")
 
-        # Test ttl
-        cache.set("foo", "bar", 10)
-        ttl = cache.ttl("foo")
-
-        if isinstance(cache.client, herd.HerdClient):
+        if isinstance(self.cache.client, herd.HerdClient):
             self.assertAlmostEqual(ttl, 12)
         else:
             self.assertAlmostEqual(ttl, 10)
 
         # Test ttl None
-        cache.set("foo", "foo", timeout=None)
-        ttl = cache.ttl("foo")
+        self.cache.set("foo", "foo", timeout=None)
+        ttl = self.cache.ttl("foo")
         self.assertIsNone(ttl)
 
         # Test ttl with expired key
-        cache.set("foo", "foo", timeout=-1)
-        ttl = cache.ttl("foo")
+        self.cache.set("foo", "foo", timeout=-1)
+        ttl = self.cache.ttl("foo")
         self.assertEqual(ttl, 0)
 
         # Test ttl with not existent key
-        ttl = cache.ttl("not-existent-key")
+        ttl = self.cache.ttl("not-existent-key")
+        self.assertEqual(ttl, 0)
+
+    def test_pttl(self):
+
+        # Test pttl
+        self.cache.set("foo", "bar", 10)
+        ttl = self.cache.pttl("foo")
+
+        # delta is set to 10 as precision error causes tests to fail
+        if isinstance(self.cache.client, herd.HerdClient):
+            self.assertAlmostEqual(ttl, 12000, delta=10)
+        else:
+            self.assertAlmostEqual(ttl, 10000, delta=10)
+
+        # Test pttl with float value
+        self.cache.set("foo", "bar", 5.5)
+        ttl = self.cache.pttl("foo")
+
+        if isinstance(cache.client, herd.HerdClient):
+            self.assertAlmostEqual(ttl, 7500, delta=10)
+        else:
+            self.assertAlmostEqual(ttl, 5500, delta=10)
+
+        # Test pttl None
+        self.cache.set("foo", "foo", timeout=None)
+        ttl = self.cache.pttl("foo")
+        self.assertIsNone(ttl)
+
+        # Test pttl with expired key
+        self.cache.set("foo", "foo", timeout=-1)
+        ttl = self.cache.pttl("foo")
+        self.assertEqual(ttl, 0)
+
+        # Test pttl with not existent key
+        ttl = self.cache.pttl("not-existent-key")
         self.assertEqual(ttl, 0)
 
     def test_persist(self):
@@ -605,6 +696,36 @@ class DjangoRedisCacheTests(unittest.TestCase):
         ttl = self.cache.ttl("foo")
         self.assertAlmostEqual(ttl, 20)
 
+    def test_pexpire(self):
+        self.cache.set("foo", "bar", timeout=None)
+        self.cache.pexpire("foo", 20500)
+        ttl = self.cache.pttl("foo")
+        # delta is set to 10 as precision error causes tests to fail
+        self.assertAlmostEqual(ttl, 20500, delta=10)
+
+    def test_expire_at(self):
+
+        # Test settings expiration time 1 hour ahead by datetime.
+        self.cache.set("foo", "bar", timeout=None)
+        expiration_time = datetime.datetime.now() + timedelta(hours=1)
+        self.cache.expire_at("foo", expiration_time)
+        ttl = self.cache.ttl("foo")
+        self.assertAlmostEqual(ttl, timedelta(hours=1).total_seconds(), delta=1)
+
+        # Test settings expiration time 1 hour ahead by Unix timestamp.
+        self.cache.set("foo", "bar", timeout=None)
+        expiration_time = datetime.datetime.now() + timedelta(hours=2)
+        self.cache.expire_at("foo", int(expiration_time.timestamp()))
+        ttl = self.cache.ttl("foo")
+        self.assertAlmostEqual(ttl, timedelta(hours=1).total_seconds() * 2, delta=1)
+
+        # Test settings expiration time 1 hour in past, which effectively
+        # deletes the key.
+        expiration_time = datetime.datetime.now() - timedelta(hours=2)
+        self.cache.expire_at("foo", expiration_time)
+        value = self.cache.get("foo")
+        self.assertIsNone(value)
+
     def test_lock(self):
         lock = self.cache.lock("foobar")
         lock.acquire(blocking=True)
@@ -647,7 +768,7 @@ class DjangoRedisCacheTests(unittest.TestCase):
         result = cache.iter_keys("foo*")
         self.assertNotEqual(next(result), None)
 
-    def test_master_slave_switching(self):
+    def test_primary_replica_switching(self):
         if isinstance(self.cache.client, ShardClient):
             self.skipTest("ShardClient doesn't support get_client")
 
@@ -703,6 +824,14 @@ class DjangoRedisCacheTests(unittest.TestCase):
         time.sleep(2)
         self.assertEqual(self.cache.get("test_key"), "foo")
 
+    def test_clear(self):
+        self.cache.set("foo", "bar")
+        value_from_cache = self.cache.get("foo")
+        self.assertEqual(value_from_cache, "bar")
+        self.cache.clear()
+        value_from_cache_after_clear = self.cache.get("foo")
+        self.assertIsNone(value_from_cache_after_clear)
+
 
 class DjangoOmitExceptionsTests(unittest.TestCase):
     def setUp(self):
@@ -760,7 +889,7 @@ class DjangoOmitExceptionsPriority2Tests(unittest.TestCase):
 
 
 # Copied from Django's sessions test suite. Keep in sync with upstream.
-# https://github.com/django/django/blob/master/tests/sessions_tests/tests.py
+# https://github.com/django/django/blob/main/tests/sessions_tests/tests.py
 class SessionTestsMixin:
     # This does not inherit from TestCase to avoid any tests being run with this
     # class, which wouldn't work, and to allow different TestCase subclasses to
@@ -1118,6 +1247,43 @@ class SessionTests(SessionTestsMixin, unittest.TestCase):
         super().test_actual_expiry()
 
 
+class TestClientClose(unittest.TestCase):
+    def setUp(self):
+        self.client = caches[DEFAULT_CACHE_ALIAS].client
+        self.client.set("TestClientClose", 0)
+
+    def tearDown(self):
+        self.client.delete("TestClientClose")
+        self.client.clear()
+
+    def test_close_client_disconnect_default(self):
+        with patch.object(self.client.connection_factory, "disconnect") as mock:
+            self.client.close()
+            assert not mock.called
+
+    @override_settings(DJANGO_REDIS_CLOSE_CONNECTION=True)
+    def test_close_disconnect_settings(self):
+        with patch.object(self.client.connection_factory, "disconnect") as mock:
+            self.client.close()
+            assert mock.called
+
+    def test_close_disconnect_settings_cache(self):
+        settings.CACHES[DEFAULT_CACHE_ALIAS]["OPTIONS"]["CLOSE_CONNECTION"] = True
+        with override_settings(CACHES=settings.CACHES):
+            # enabling override_settings context emits 'setting_changed' signal
+            # (re-set the value to populate again client connections)
+            self.client.set("TestClientClose", 0)
+            with patch.object(self.client.connection_factory, "disconnect") as mock:
+                self.client.close()
+                assert mock.called
+
+    def test_close_disconnect_client_options(self):
+        self.client._options["CLOSE_CONNECTION"] = True
+        with patch.object(self.client.connection_factory, "disconnect") as mock:
+            self.client.close()
+            assert mock.called
+
+
 class TestDefaultClient(unittest.TestCase):
     @patch("test_backend.DefaultClient.get_client")
     @patch("test_backend.DefaultClient.__init__", return_value=None)
